<?php

namespace QUI\Verification;

abstract class AbstractVerification implements VerificationInterface
{
    /**
     * Verification identifier
     *
     * @var string|int
     */
    protected string|int $identifier;

    /**
     * Additional data
     *
     * @var array
     */
    protected array $additionalData = [];

    /**
     * VerificationInterface constructor.
     *
     * @param int|string $identifier - Unique identifier
     * @param array $additionalData (optional) - Additional data
     */
    public function __construct(int|string $identifier, array $additionalData = [])
    {
        $this->identifier = $identifier;
        $this->additionalData = $additionalData;
    }

    public function setIdentifier(int|string $identifier): void
    {
        $this->identifier = $identifier;
    }

    /**
     * Get an identifier of this Verification
     *
     * @return string|int
     */
    public function getIdentifier(): int|string
    {
        return $this->identifier;
    }

    /**
     * Get additional data
     *
     * @return array
     */
    public function getAdditionalData(): array
    {
        return $this->additionalData;
    }

    public function setAdditionalData(array $data): void
    {
        $this->additionalData = $data;
    }

    /**
     * Get the duration of a Verification (minutes)
     *
     * @return int|false - duration in minutes;
     * if this method returns false use the module setting default value
     */
    public function getValidDuration(): bool|int
    {
        return false;
    }

    /**
     * Execute this method on successful verification
     *
     * @return void
     */
    abstract public function onSuccess(): void;

    /**
     * Execute this method on unsuccessful verification
     *
     * @return void
     */
    abstract public function onError(): void;

    /**
     * This message is displayed to the user on successful verification
     *
     * @return string
     */
    abstract public function getSuccessMessage(): string;

    /**
     * This message is displayed to the user on unsuccessful verification
     *
     * @param string $reason - The reason for the error (see \QUI\Verification\Verifier::REASON_)
     * @return string
     */
    abstract public function getErrorMessage(string $reason): string;

    /**
     * Automatically redirect the user to this URL on successful verification
     *
     * @return string|false - If this method returns false, no redirection takes place
     */
    public function getOnSuccessRedirectUrl(): bool|string
    {
        return false;
    }

    /**
     * Automatically redirect the user to this URL on unsuccessful verification
     *
     * @return string|false - If this method returns false, no redirection takes place
     */
    public function getOnErrorRedirectUrl(): bool|string
    {
        return false;
    }

    /**
     * Get type (class name)
     *
     * @return string
     */
    public static function getType(): string
    {
        return get_called_class();
    }
}
